 ! MODULE GENERAL_PURPOSE
 ! general methods for integration and matching
 module general_purpose
   implicit none

 contains

   ! PREDICTOR CORRECTOR subroutine
   ! Predictor corrector subroutine for our system of equations; f is the
   ! function, df its derivative, pot the potential, energy the trial energy, deltax
   ! the step on the x axis; initial conditions must be defined on f(0) and df(0).
   subroutine predictor_corrector (f, df, pot, deltax, energy)
     implicit none
     
     real(kind=8), intent(inout)::f(0:),df(0:)
     real(kind=8), intent(in)   ::pot(0:), deltax, energy
     integer                    ::i
       
     do i=1,size(f)-1
         
       df(i)=df(i-1)+deltax*(pot(i-1)-dble(2)*energy)*f(i-1)
       f(i) = f(i-1) + deltax*df(i-1)
       
       df(i)=df(i-1)+0.5d0*deltax*((pot(i-1)-dble(2)*energy)*f(i-1)+(pot(i)-dble(2)*energy)*f(i))
       f(i) = f(i-1) + 0.5d0*deltax*(df(i-1)+df(i))
         
     enddo
   endsubroutine
   
   ! RETURN_DELTAZ function 
   ! match the right slice of function and derivative arrays with the left
   ! slice at the matching point of index p_match
   function return_deltaz(f, df, pot, deltax, energy, p_match)
     implicit none
     
     real(kind=8), intent(inout)::f(0:),df(0:)
     real(kind=8), intent(in)   ::pot(0:), deltax, energy
     integer, intent(in)        ::p_match
     real(kind=8)               ::return_deltaz
     
     real(kind=8), allocatable::fl,dfl,fr,dfr
     integer                  ::N
     integer                  ::i
     
     N=size(f)

     call predictor_corrector(f(0:p_match),df(0:p_match),pot(0:p_match),deltax,energy)
 
     ! store function and derivative value in the matching point obtained by
     !  integrating from the left
     fl = f(p_match)
     dfl = df(p_match)
     
     ! integration in the right slice
     call predictor_corrector(f(N-1:p_match:-1),df(N-1:p_match:-1),pot(N-1:p_match:-1),deltax,energy)

     ! store function value in the matching point obtained after integrating
     ! from the right
     fr = f(p_match)     

     ! right slice resizing
     f(p_match:N-1)=(fl/fr)*f(p_match:N-1)
     df(p_match:N-1)=(fl/fr)*(-df(p_match:N-1)) ! minus because of the direction of integration
     dfr = df(p_match)                          ! store derivative value in the matching point after resizing
     
     return_deltaz = (dfl-dfr)**2 ! difference between derivatives in the
                                  ! matching point we want to find the minimum of
     return
   endfunction 
   
   ! INTEGRAL function
   function integral(f,deltax)
     implicit none

     real(kind=8),intent(in)::f(0:),deltax
     integer                ::N,i
     real(kind=8)           ::integral
     
     N=size(f)
     integral=0
     do i=1,N-1
       integral=integral+0.5d0*(f(i)+f(i-1))*deltax
     enddo
     return
   endfunction integral

 endmodule general_purpose

 ! MODULE COMM
 ! contains data and subroutines used by the main program
 module comm
   use general_purpose
   implicit none
   
   ! x axis, function and derivative grids
   integer                  ::N
   real(kind=8)             ::h
   real(kind=8), allocatable::x(:), pot(:), y(:,:), z(:,:)
   
   ! matching variables
   integer                  ::i_match
   real(kind=8), allocatable::deltaz(:)
   
   ! energy computation variables
   real(kind=8)             ::tol, deriv_min, eig_max, deltaE
   integer                  ::M, L, it_max, L_a
   real(kind=8), allocatable::E(:)   ! trial energies
   integer, allocatable     ::i_E(:) ! indices of the approximated minima

   ! final eigenvalues array
   real(kind=8),allocatable::E_out(:)

   ! indexes
   integer, private::i,j,k
 
 contains
  
   ! INIT subroutine
   subroutine init 
     implicit none

     ! read x-grid and potential data
     open (unit=1, file='input/pot.dat')
     read (1,*) N
     allocate(x(0:N-1), pot(0:N-1))
     do i=0,N-1
       read(1, *) x(i), pot(i)
     enddo
     close(1)
     
     ! load eigenvalue computation parameters
     open (unit=1, file='input/eigenvalue_input.dat')
     read(1, *) tol
     read(1, *) i_match
     read(1, *) eig_max
     read(1, *) M
     read(1, *) L
     read(1, *) it_max
     close(1)
     
     ! use loaded data for array allocation
     allocate(y(0:N-1,0:M-1), z(0:N-1,0:M-1))
     allocate(deltaz(0:M-1))
     allocate(E(0:M-1),i_E(0:L-1))

     ! x axis step
     h=x(1)-x(0)      

     ! energy grid
     deltaE=dble(eig_max)/dble(M)
     do i=0, M-1
       E(i)=i*deltaE
     enddo
   endsubroutine init
   
   ! FILL_DELTAZ subroutine
   subroutine fill_deltaz
     implicit none

     ! compute the differences between derivatives we want to find the minimum of
     do i=0,M-1
       deltaz(i) = return_deltaz(y(:,i),z(:,i),pot,h,E(i),i_match)
     enddo
   endsubroutine fill_deltaz
   
   ! BRACKET_MINIMA subroutine
   subroutine bracket_minima
     implicit none

     ! minima bracketing
     L_a=0
     do i=1, M-2
       if (deltaz(i)<deltaz(i+1).and.deltaz(i)<deltaz(i-1)) then
         i_E(L_a)=i
         L_a=L_a+1
       endif
       if (L_a==L) then
         exit
       endif
     enddo
     ! allocate the array with the final eigenvalues
     allocate(E_out(0:L_a-1))
   endsubroutine bracket_minima
   
   ! GOLDEN SECTION SEARCH subroutine
   subroutine golden_section_search
     implicit none

     real(kind=8), parameter::C=0.5d0*(3.d0-sqrt(5.d0)), R=1.d0-C
     real(kind=8)           ::f1, f2
     real(kind=8)           ::ax, bx, cx, x0, x1, x2, x3

     do i=0, L_a-1
       ax = E(i_E(i)-1)
       bx = E(i_E(i))
       cx = E(i_E(i)+1)
       x0=ax       
       x3=cx
       if(abs(cx-bx)>abs(bx-ax)) then
         x1=bx
         x2=bx+C*(cx-bx)
       else
         x2=bx
         x1=bx-C*(bx-ax)
       endif 
       f1=return_deltaz(y(:,i_E(i)),z(:,i_E(i)),pot,h,x1,i_match)
       f2=return_deltaz(y(:,i_E(i)),z(:,i_E(i)),pot,h,x2,i_match)
       j=0
       do while(abs(x3-x0)>tol*(abs(x1)+abs(x2)))
         if (f2<f1) then
           x0=x1
           x1=x2
           x2=R*x1+C*x3
           f1=f2
           f2=return_deltaz(y(:,i_E(i)),z(:,i_E(i)),pot,h,x2,i_match)
         else
           x3=x2
           x2=x1
           x1=R*x2+C*x0
           f2=f1
           f1=return_deltaz(y(:,i_E(i)),z(:,i_E(i)),pot,h,x1,i_match)
         endif
         j=j+1
         if (j==it_max) then
           print *, 'Reached maximum number of iterations for eigenvalue number ', i
           exit
         endif
       enddo

       if(f1>f2) then
         E_out(i)=x1
       else
         E_out(i)=x2
       endif
       ! print *, i, 'golden ratio', (x1-x0)/(x3-x0)
     enddo
   endsubroutine golden_section_search
   
   ! NORMALIZE_EIGENFUNCTIONS subroutine
   subroutine normalize_eigenfunctions
   implicit none
   
   real(kind=8)::factor

   do i=0,L_a-1
    factor=integral(y(:,i_E(i))**2,h)
    y(:,i_E(i))=y(:,i_E(i))/sqrt(factor)
   enddo
   endsubroutine normalize_eigenfunctions
   
! SUBROUTINES FOR PRINTNG

   ! PRINT_DELTAZ subroutine
   subroutine print_deltaz
     implicit none

     open (unit=1, file='output/deltaz.dat')
     do i=0, M-1
       write(1, *) E(i), deltaz(i)
     enddo
     close(1)
   endsubroutine print_deltaz
  
   ! PRINT_EIGENVALUES subroutine
   subroutine print_eigenvalues
     implicit none

     open (unit=1, file='output/eigenvalues.dat')
     write(1, *) 0, E_out(0)    
     do i=1, L_a-1
       write(1, *) i, E_out(i), E_out(i)-E_out(i-1)
     enddo
   endsubroutine print_eigenvalues
   
   ! PRINT_EIGENFUNCTIONS subroutine
   subroutine print_eigenfunctions(nfuncs)
     implicit none

     integer::nfuncs

     open (unit=1, file='output/eigenfunctions.dat')
     do i=0, N-1
       write(1, *) x(i), y(i,i_E(0:nfuncs-1))
     enddo
   endsubroutine print_eigenfunctions

 endmodule comm


 ! MAIN PROGRAM
 program main
   use comm
   implicit none

   integer::i

   ! INIT 
   ! load potential and grid parameters
   print *, 'Loading input data and allocating arrays...'
   call init
    
   ! INTIAL CONDITIONS FOR PREDICTOR CORRECTOR
   print *, 'Setting initial conditions...'
   y(0,:)=0.d0
   z(0,:)=0.005d0
   y(N-1,:)=0.d0
   z(N-1,:)=0.005d0
   
   ! COMPUTE THE DELTAZ AT ALL THE ENERGIES
   print *, 'Computing the derivative differences at the matching point...'
   call fill_deltaz

   ! MINIMA BRACKETING
   print *, 'Minima bracketing...'
   call bracket_minima
   if (L_a/=L) then
     print *, L_a, ' eigenvalues have been found.'
   else
     print *, 'Max number of eigenvalues reached.'
   endif
   print *, 'Found eigenvalues at approximately E='
   do i=0,L_a-1
     print *, E(i_E(i))
   enddo

   ! DELTAZ OUTPUT
   ! call print_deltaz 
   
   ! GOLDEN SECTION SEARCH OF MINIMA 
   print *, 'Golden section search of minima...'
   call golden_section_search

   ! EIGENFUNCTIONS NORMALIZATION
   print *, 'Normalizing eigenfunctions...'
   call normalize_eigenfunctions

   ! PRINT EIGENVALUES
   print *, 'Writing results to output files...'
   call print_eigenvalues
   call print_eigenfunctions(L_a)
   close(1)
   
 endprogram main
